# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

from datetime import datetime
import logging
import os
import pathlib
from unittest import mock

import pytest

from ... import errors, model
from ...components.core import SimpleRepository
from ...components.resource_cache import ResourceCacheRepository
from .fake_repository import FakeRepository


@pytest.fixture
def repository(tmp_path: pathlib.Path) -> ResourceCacheRepository:
    http_resource = model.HttpResource("url/http-1.0-any.whl")
    http_resource.context["etag"] = "etag"

    not_to_cache_resource = model.HttpResource("url/uncacheable-1.0-any.whl", to_cache=False)
    not_to_cache_resource.context["etag"] = "etag"
    source = FakeRepository(
        resources={
            "http-1.0-any.whl": http_resource,
            "local-1.0.tar.gz": model.LocalResource(pathlib.Path("path")),
            "http_no_etag-1.0-any.whl": model.HttpResource("url/http_no_etag-1.0-any.whl"),
            "uncacheable-1.0-any.whl": not_to_cache_resource,
        },
    )
    return ResourceCacheRepository(
        source=source,
        cache_path=tmp_path,
        http_client=mock.MagicMock(),
    )


@pytest.mark.asyncio
async def test_get_resource__cache_hit(repository: ResourceCacheRepository) -> None:
    (repository._cache_path / "http").mkdir()
    cached_file = repository._cache_path / "http" / "http-1.0-any.whl"
    cached_file.write_text("cached content")
    cached_info_file = repository._cache_path / "http" / "http-1.0-any.whl.info"
    # The content of the the info file matches the upstream cache
    cached_info_file.write_text("etag")

    context = model.RequestContext(repository)
    resource = await repository.get_resource(
        project_name="http",
        resource_name="http-1.0-any.whl",
        request_context=context,
    )
    # The cache returns a LocalResource pointing to
    # the cached file with the same etag as upstream
    assert isinstance(resource, model.LocalResource)
    assert resource.path == cached_file
    assert resource.context["etag"] == "etag"
    assert cached_file.read_text() == "cached content"


@pytest.mark.asyncio
async def test_get_resource__cache_miss__wrong_etag(repository: ResourceCacheRepository) -> None:
    (repository._cache_path / "http").mkdir()
    cached_file = repository._cache_path / "http" / "http-1.0-any.whl"
    cached_file.write_text("invalid cached content")
    cached_info_file = repository._cache_path / "http" / "http-1.0-any.whl.info"
    # The content of the the info file matches the upstream cache
    cached_info_file.write_text("wrong_etag")

    context = model.RequestContext(repository)

    with mock.patch(
        "simple_repository.utils.download_file",
        mock.AsyncMock(
            side_effect=lambda **kwargs: kwargs["dest_file"].touch(),
        ),
    ):
        response = await repository.get_resource(
            project_name="http",
            resource_name="http-1.0-any.whl",
            request_context=context,
        )

    # after the cache miss, the upstream file is downloaded and the info file is created.
    # The cache returns a LocalResource pointing to the downloaded file.
    assert isinstance(response, model.LocalResource)
    assert response.path == repository._cache_path / "http" / "http-1.0-any.whl"
    assert (repository._cache_path / "http" / "http-1.0-any.whl.info").read_text() == "etag"
    assert (repository._cache_path / "http" / "http-1.0-any.whl").read_text() != "invalid cached content"


@pytest.mark.asyncio
async def test_get_resource__cache_miss__no_etag(repository: ResourceCacheRepository) -> None:
    assert not (repository._cache_path / "http" / "http-1.0-any.whl.info").is_file()
    assert not (repository._cache_path / "http" / "http-1.0-any.whl").is_file()
    context = model.RequestContext(repository)

    with mock.patch(
        "simple_repository.utils.download_file",
        mock.AsyncMock(
            side_effect=lambda **kwargs: kwargs["dest_file"].touch(),
        ),
    ):
        response = await repository.get_resource(
            project_name="http",
            resource_name="http-1.0-any.whl",
            request_context=context,
        )

    # after the cache miss, the upstream file is downloaded and the info file is created.
    # The cache returns a LocalResource pointing to the downloaded file.
    assert isinstance(response, model.LocalResource)
    assert response.path == repository._cache_path / "http" / "http-1.0-any.whl"
    assert (repository._cache_path / "http" / "http-1.0-any.whl.info").is_file()
    assert (repository._cache_path / "http" / "http-1.0-any.whl").is_file()
    assert response.context.get("etag") == "etag"


@pytest.mark.asyncio
async def test_get_resource__cache_miss_local_resource(
    repository: ResourceCacheRepository,
) -> None:
    context = model.RequestContext(repository)
    resource = await repository.get_resource(
        project_name="local",
        resource_name="local-1.0.tar.gz",
        request_context=context,
    )

    # the cache doesn't store LocalResources. No cache file or info file is created.
    assert isinstance(resource, model.LocalResource)
    assert resource.path == pathlib.Path("path")
    assert not (repository._cache_path / "local" / "local-1.0.tar.gz.info").is_file()
    assert not (repository._cache_path / "local" / "local-1.0.tar.gz").is_file()


@pytest.mark.asyncio
async def test_get_resource__path_traversal(
    repository: ResourceCacheRepository,
) -> None:
    context = model.RequestContext(repository)
    with pytest.raises(
        ValueError,
        match="is not contained in",
    ):
        await repository.get_resource(
            project_name="not-used",
            resource_name="../../../etc/passwords",
            request_context=context,
        )


@pytest.mark.asyncio
async def test_get_resource__source_unavailable_cache_hit(
    tmp_path: pathlib.Path,
) -> None:
    source = mock.AsyncMock(
        spec=SimpleRepository,
        get_resource=mock.AsyncMock(side_effect=errors.SourceRepositoryUnavailable),
    )

    (tmp_path / "project-name").mkdir()

    (tmp_path / "project-name" / "resource-name.info").write_text("etag")
    (tmp_path / "project-name" / "resource-name").write_text("content")

    repository = ResourceCacheRepository(
        source=source,
        cache_path=tmp_path,
        http_client=mock.MagicMock(),
    )

    resource = await repository.get_resource("project-name", "resource-name")
    assert isinstance(resource, model.LocalResource)
    assert resource.path == tmp_path / "project-name" / "resource-name"


@pytest.mark.asyncio
async def test_get_resource__source_unavailable_cache_hit__falback_disabled(
    tmp_path: pathlib.Path,
) -> None:
    source = mock.AsyncMock(
        spec=SimpleRepository,
        get_resource=mock.AsyncMock(side_effect=errors.SourceRepositoryUnavailable),
    )

    (tmp_path / "project-name").mkdir()

    (tmp_path / "project-name" / "resource-name.info").write_text("etag")
    (tmp_path / "project-name" / "resource-name").write_text("content")

    repository = ResourceCacheRepository(
        source=source,
        cache_path=tmp_path,
        http_client=mock.MagicMock(),
        fallback_to_cache=False,
    )

    with pytest.raises(errors.SourceRepositoryUnavailable):
        await repository.get_resource("project-name", "resource-name")


@pytest.mark.asyncio
async def test_get_resource__source_unavailable_cache_hit__log(
    tmp_path: pathlib.Path,
) -> None:
    source = mock.AsyncMock(
        spec=SimpleRepository,
        get_resource=mock.AsyncMock(side_effect=errors.SourceRepositoryUnavailable),
    )

    (tmp_path / "project-name").mkdir()

    (tmp_path / "project-name" / "resource-name.info").write_text("etag")
    (tmp_path / "project-name" / "resource-name").write_text("content")

    mock_logger = mock.Mock(spec=logging.Logger)

    repository = ResourceCacheRepository(
        source=source,
        cache_path=tmp_path,
        http_client=mock.MagicMock(),
        logger=mock_logger,
    )

    await repository.get_resource("project-name", "resource-name")
    mock_logger.error.assert_called_once_with(
        "Upstream unavailable, served cached project-name:resource-name",
    )


@pytest.mark.asyncio
async def test_get_resource__source_unavailable_cache_miss(
    tmp_path: pathlib.Path,
) -> None:
    source = mock.AsyncMock(
        spec=SimpleRepository,
        get_resource=mock.AsyncMock(side_effect=errors.SourceRepositoryUnavailable),
    )

    repository = ResourceCacheRepository(
        source=source,
        cache_path=tmp_path,
        http_client=mock.MagicMock(),
    )

    with pytest.raises(errors.SourceRepositoryUnavailable):
        await repository.get_resource("project-name", "resource-name")


@pytest.mark.asyncio
async def test_get_resource__no_cache_created_when_no_upstream_etag_exists(
    repository: ResourceCacheRepository,
) -> None:
    context = model.RequestContext(repository)
    resource = await repository.get_resource(
        project_name="http-no-etag",
        resource_name="http_no_etag-1.0-any.whl",
        request_context=context,
    )

    # Upstream doesn't set an etag so the HttpResource is not cached,
    # otherwise a LocalResource would have been returned
    # No cache file or info file gets created and an http resource is returned.
    assert isinstance(resource, model.HttpResource)
    assert not (repository._cache_path / "http-no-etag" / "http_no_etag-1.0-any.whl").is_file()
    assert not (repository._cache_path / "http-no-etag" / "http_no_etag-1.0-any.whl.info").is_file()


@pytest.mark.parametrize(
    "resource", [
        model.HttpResource(url="url", to_cache=False),
        model.LocalResource(path=pathlib.Path("."), to_cache=False),
        model.TextResource(text="text", to_cache=False),
    ],
)
@pytest.mark.asyncio
async def test_get_resource__no_cache_created_when_to_cache_is_false(
    resource: model.Resource,
    tmp_path: pathlib.Path,
) -> None:
    resource.context["etag"] = "etag"

    repository = ResourceCacheRepository(
        source=FakeRepository(resources={"resource-1.0-any.whl": resource}),
        cache_path=tmp_path,
        http_client=mock.MagicMock(),
    )

    context = model.RequestContext(repository)
    result = await repository.get_resource(
        project_name="resource",
        resource_name="resource-1.0-any.whl",
        request_context=context,
    )

    # Upstream sets an etag but the to_cache attribute is False so the resource is not cached.
    # No cache file or info file gets created and an same resource is returned.
    assert resource == result
    assert not (repository._cache_path / "resource" / "resource-1.0-any.whl").is_file()
    assert not (repository._cache_path / "resource" / "resource-1.0-any.whl.info").is_file()


@pytest.mark.asyncio
async def test_get_resource__source_raised_not_modified__request_etag_invalid(
    repository: ResourceCacheRepository,
) -> None:
    # Upsream raised not modified and the request context misses the etag.
    repository.source = mock.Mock(get_resource=mock.AsyncMock(side_effect=model.NotModified))

    (repository._cache_path / "http").mkdir()
    cached_file = repository._cache_path / "http" / "http-1.0-any.whl"
    cached_file.write_text("cached content")
    cached_info_file = repository._cache_path / "http" / "http-1.0-any.whl.info"
    # The content of the the info file matches the upstream cache
    cached_info_file.write_text("etag")

    context = model.RequestContext(repository)
    resource = await repository.get_resource(
        project_name="http",
        resource_name="http-1.0-any.whl",
        request_context=context,
    )
    # The cache returns a LocalResource pointing to
    # the cached file with the same etag as upstream
    assert isinstance(resource, model.LocalResource)
    assert resource.path == cached_file
    assert resource.context["etag"] == "etag"
    assert cached_file.read_text() == "cached content"


@pytest.mark.asyncio
async def test_get_resource__source_raised_not_modified__request_etag_valid(
    repository: ResourceCacheRepository,
) -> None:
    # Upsream raised not modified and the request etag matches.
    repository.source = mock.Mock(get_resource=mock.AsyncMock(side_effect=model.NotModified))

    (repository._cache_path / "http").mkdir()
    cached_file = repository._cache_path / "http" / "http-1.0-any.whl"
    cached_file.write_text("cached content")
    cached_info_file = repository._cache_path / "http" / "http-1.0-any.whl.info"
    # The content of the the info file matches the upstream cache
    cached_info_file.write_text("etag")

    context = model.RequestContext(repository, {"etag": "etag"})

    with pytest.raises(model.NotModified):
        await repository.get_resource(
            project_name="http",
            resource_name="http-1.0-any.whl",
            request_context=context,
        )


def test_resource_cache_init(tmp_path: pathlib.Path) -> None:
    real_repo = tmp_path / "dir" / "cache_repo"
    real_repo.mkdir(parents=True)

    symlink = tmp_path / "link"
    symlink.symlink_to(real_repo)

    repo = ResourceCacheRepository(
        source=mock.AsyncMock(),
        cache_path=symlink,
        http_client=mock.MagicMock(),
    )
    assert str(symlink) != str(real_repo)
    assert str(repo._cache_path) == str(real_repo)


def test_update_last_access(repository: ResourceCacheRepository) -> None:
    cached_info = repository._cache_path / "my_resource.ext.info"
    cached_info.touch()

    with mock.patch(
        "datetime.datetime",
        mock.Mock(
            now=mock.Mock(return_value=datetime.fromisoformat("2006-07-09")),
            fromisoformat=datetime.fromisoformat,
            spec=datetime,
        ),
    ):
        repository._update_last_access(cached_info)

    assert os.path.getatime(cached_info) == datetime.fromisoformat("2006-07-09").timestamp()

    with mock.patch(
        "datetime.datetime",
        mock.Mock(
            now=mock.Mock(return_value=datetime.fromisoformat("2025-07-09")),
            fromisoformat=datetime.fromisoformat,
            spec=datetime,
        ),
    ):
        repository._update_last_access(cached_info)

    assert os.path.getatime(cached_info) == datetime.fromisoformat("2025-07-09").timestamp()


@pytest.mark.asyncio
async def test_update_last_access__cache_hit_called(
    repository: ResourceCacheRepository,
) -> None:
    (repository._cache_path / "http").mkdir()
    cached_file = (repository._cache_path / "http" / "http-1.0-any.whl")
    cached_file.touch()
    cached_info_file = repository._cache_path / "http" / "http-1.0-any.whl.info"
    cached_info_file.write_text("etag")

    update_last_access_mock = mock.Mock()

    context = model.RequestContext(repository)
    with mock.patch.object(
        target=ResourceCacheRepository,
        attribute="_update_last_access",
        new=update_last_access_mock,
    ):
        await repository.get_resource(
            project_name="http",
            resource_name="http-1.0-any.whl",
            request_context=context,
        )

    update_last_access_mock.assert_called_once()


@pytest.mark.asyncio
async def test_update_last_access_for__cache_miss_local_not_called(
    repository: ResourceCacheRepository,
) -> None:
    update_last_access_for_mock = mock.Mock()
    context = model.RequestContext(repository)
    with mock.patch.object(
        target=ResourceCacheRepository,
        attribute="_update_last_access",
        new=update_last_access_for_mock,
    ):
        await repository.get_resource(
            project_name="local",
            resource_name="local-1.0.tar.gz",
            request_context=context,
        )
    update_last_access_for_mock.assert_not_called()


@pytest.mark.asyncio
async def test_update_last_access_for__cache_miss_remote_called(
    repository: ResourceCacheRepository,
) -> None:
    context = model.RequestContext(repository)
    update_last_access_for_mock = mock.Mock()
    with mock.patch(
        "simple_repository.utils.download_file",
        mock.AsyncMock(
            side_effect=lambda **kwargs: kwargs["dest_file"].touch(),
        ),
    ), mock.patch.object(
        target=ResourceCacheRepository,
        attribute="_update_last_access",
        new=update_last_access_for_mock,
    ):
        await repository.get_resource(
            project_name="http",
            resource_name="http-1.0-any.whl",
            request_context=context,
        )
    update_last_access_for_mock.assert_called_once()


@pytest.mark.asyncio
async def test_store_resource__http_resource(
    repository: ResourceCacheRepository,
    tmp_path: pathlib.Path,
) -> None:
    http_resource = model.HttpResource(url="my_url")
    cache_path = tmp_path / "cache"

    with mock.patch(
        "simple_repository.utils.download_file",
        mock.AsyncMock(
            side_effect=lambda **kwargs: kwargs["dest_file"].write_text("http content"),
        ),
    ):
        await repository._store_resource(
            resource=http_resource,
            upstream_etag="etag",
            resource_path=cache_path,
            resource_info_path=cache_path.with_suffix(".metadata"),
        )

    assert cache_path.read_text() == "http content"
    assert cache_path.with_suffix(".metadata").read_text() == "etag"


@pytest.mark.asyncio
async def test_store_resource__local_resource(
    repository: ResourceCacheRepository,
    tmp_path: pathlib.Path,
) -> None:
    file_path = tmp_path / "file"
    (tmp_path / "file").write_text("file_content")
    cache_path = tmp_path / "cache"
    local_resource = model.LocalResource(file_path)

    await repository._store_resource(
        resource=local_resource,
        upstream_etag="etag",
        resource_path=cache_path,
        resource_info_path=cache_path.with_suffix(".metadata"),
    )

    assert cache_path.read_text() == "file_content"
    assert cache_path.with_suffix(".metadata").read_text() == "etag"


@pytest.mark.asyncio
async def test_store_resource__text_resource(
    repository: ResourceCacheRepository,
    tmp_path: pathlib.Path,
) -> None:
    cache_path = tmp_path / "cache"
    text_resource = model.TextResource("my_text")

    await repository._store_resource(
        resource=text_resource,
        upstream_etag="etag",
        resource_path=cache_path,
        resource_info_path=cache_path.with_suffix(".metadata"),
    )

    assert cache_path.read_text() == "my_text"
    assert cache_path.with_suffix(".metadata").read_text() == "etag"



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

import pathlib

import packaging.utils

from .. import errors, model, utils
from .core import SimpleRepository


class AllowListedRepository(SimpleRepository):
    """Exposes only the whitelisted projects of the source repository.
    Projects available from the source but not added to the
    whitelist file are not made available from this repository.
    """
    def __init__(
        self,
        source: SimpleRepository,
        special_case_file: pathlib.Path,
    ) -> None:
        self.source = source
        self._special_cases: dict[str, str] = self._load_config_json(special_case_file)
        self.special_case_file = special_case_file

    async def get_project_page(
        self,
        project_name: str,
        *,
        request_context: model.RequestContext = model.RequestContext.DEFAULT,
    ) -> model.ProjectDetail:
        """Returns the project page from the source if the project_name
        is whitelisted.
        """
        if project_name not in self._special_cases:
            raise errors.PackageNotFoundError(project_name)
        else:
            return await self.source.get_project_page(
                project_name,
                request_context=request_context,
            )

    async def get_project_list(
        self,
        *,
        request_context: model.RequestContext = model.RequestContext.DEFAULT,
    ) -> model.ProjectList:
        return model.ProjectList(
            meta=model.Meta("1.0"),
            projects=frozenset(
                model.ProjectListElement(name) for name in
                self._special_cases.keys()
            ),
        )

    async def get_resource(
        self,
        project_name: str,
        resource_name: str,
        *,
        request_context: model.RequestContext = model.RequestContext.DEFAULT,
    ) -> model.Resource:
        if project_name in self._special_cases:
            return await self.source.get_resource(
                project_name,
                resource_name,
                request_context=request_context,
            )
        raise errors.ResourceUnavailable(resource_name)

    def _load_config_json(self, json_file: pathlib.Path) -> dict[str, str]:
        json_config = utils.load_config_json(json_file)

        config_dict: dict[str, str] = {}
        for key, value in json_config.items():
            if (
                not isinstance(key, str) or
                not isinstance(value, str)
            ):
                raise errors.InvalidConfigurationError(
                    f'Invalid special case configuration file. {json_file} '
                    'must contain a dictionary mapping a project name to a tuple'
                    ' containing a glob pattern and a yank reason.',
                )
            config_dict[packaging.utils.canonicalize_name(key)] = value

        return config_dict



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

from datetime import datetime, timedelta
from unittest import mock

import pytest

from ... import model
from ...components.new_releases_remover import NewReleasesRemover
from .fake_repository import FakeRepository


def create_project_detail(
    *dates: datetime | None,
    project_name: str = "project",
) -> model.ProjectDetail:
    files = tuple(
        model.File(
            filename=f"{project_name}-{i}.whl",
            url="url",
            hashes={},
            upload_time=date,
        ) for i, date in enumerate(dates)
    )

    return model.ProjectDetail(
        meta=model.Meta("1.0"),
        name=project_name,
        files=files,
    )


def test_exclude_recent_distributions__old_files() -> None:
    repository = NewReleasesRemover(
        source=FakeRepository(),
        quarantine_time=timedelta(days=10),
    )
    now = datetime(2023, 1, 1)
    project_detail = create_project_detail(datetime(1926, 1, 1), datetime(2000, 1, 4))
    new_project_detail = repository._exclude_recent_distributions(project_detail, now)
    assert new_project_detail == project_detail


def test_exclude_recent_distributions__new_files() -> None:
    repository = NewReleasesRemover(
        source=FakeRepository(),
        quarantine_time=timedelta(days=10),
    )

    now = datetime(2023, 1, 1)
    project_detail = create_project_detail(now, now - timedelta(days=1), now - timedelta(days=11))
    new_project_detail = repository._exclude_recent_distributions(project_detail, now)
    assert new_project_detail != project_detail
    assert len(new_project_detail.files) == 1
    assert new_project_detail.files[0].upload_time == now - timedelta(days=11)


@pytest.mark.asyncio
async def test_get_project_page() -> None:
    source = FakeRepository(
        project_pages=[
            create_project_detail(
                datetime.now(),
                datetime(1926, 1, 1),
                None,
            ),
        ],
    )
    repository = NewReleasesRemover(
        source=source,
        quarantine_time=timedelta(days=10),
    )

    mock_project_detail = mock.Mock(spec=model.ProjectDetail)
    mock_datetime = mock.Mock(spec=datetime)
    mock_datetime.now.return_value = datetime(2000, 1, 4)
    with (
        mock.patch(
            "simple_repository.components.new_releases_remover.NewReleasesRemover._exclude_recent_distributions",
            return_value=mock_project_detail,
        ) as mock_exclude_recent_distributions,
        mock.patch(
            "simple_repository.components.new_releases_remover.datetime",
            mock_datetime,
        ),
    ):
        project_page = await repository.get_project_page("project")

    source_project_page = await source.get_project_page("project")
    mock_exclude_recent_distributions.assert_called_once_with(project_page=source_project_page, now=datetime(2000, 1, 4))
    assert project_page == mock_project_detail


@pytest.mark.asyncio
async def test_whitelist() -> None:
    whitelisted_project_list = create_project_detail(
        datetime.now(),
        project_name="project1",
    )

    regular_project_list = create_project_detail(
        datetime.now(),
        datetime(1926, 1, 1),
        datetime(2000, 1, 4),
        project_name="project2",
    )

    repository = NewReleasesRemover(
        source=FakeRepository(
            project_pages=[
                whitelisted_project_list,
                regular_project_list,
            ],
        ),
        whitelist=("project1", "another-project"),
    )

    project_list = await repository.get_project_page("project2")
    assert project_list != regular_project_list
    assert len(project_list.files) == 2

    project_list = await repository.get_project_page("project1")
    assert project_list == whitelisted_project_list



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

from datetime import datetime
import json
from typing import Optional, Union

import pytest

from ..model import File, Meta, ProjectDetail, ProjectList, ProjectListElement
from ..serializer import SerializerHtmlV1, SerializerJsonV1


def test_serialize_file_html() -> None:
    serializer = SerializerHtmlV1()

    file = File(
        filename="test.html",
        url="https://example.com/test.html",
        hashes={"test": "123", "sha256": "abc123"},
        requires_python=">=3.6",
    )
    expected = (
        '<a href="https://example.com/test.html#sha256=abc123" data-requires-python="&gt;=3.6"'
        '>test.html</a><br/>\n'
    )
    assert serializer._serialize_file(file) == expected

    file = File(
        filename="test.html",
        url="https://example.com/test.html",
        hashes={},
    )
    expected = (
        '<a href="https://example.com/test.html"'
        '>test.html</a><br/>\n'
    )
    assert serializer._serialize_file(file) == expected


@pytest.mark.parametrize(
    "yank_attr, yank_value",
    [
        (' data-yanked=""', True),
        (' data-yanked="reason"', "reason"),
        ('', None),
        ('', False),
    ],
)
def test_serialize_file_html_yank(
    yank_attr: str,
    yank_value: Optional[Union[bool, str]],
) -> None:
    serializer = SerializerHtmlV1(

    )
    file = File(
        filename="test.html",
        url="https://example.com/test.html",
        hashes={},
        yanked=yank_value,
    )
    expected = (
        '<a href="https://example.com/test.html"'
        f'{yank_attr}'
        '>test.html</a><br/>\n'
    )
    assert serializer._serialize_file(file) == expected


@pytest.mark.parametrize(
    "metadata_attr, metadata_value",
    [
        (' data-core-metadata="true"', True),
        (' data-core-metadata="sha=..."', {"sha": "..."}),
        ('', None),
        ('', False),
    ],
)
def test_serialize_file_html_metadata(
    metadata_attr: str,
    metadata_value: Optional[Union[bool, dict[str, str]]],
) -> None:
    serializer = SerializerHtmlV1()

    file = File(
        filename="test.html",
        url="https://example.com/test.html",
        hashes={},
        dist_info_metadata=metadata_value,
    )
    expected = (
        '<a href="https://example.com/test.html"'
        f'{metadata_attr}'
        '>test.html</a><br/>\n'
    )
    assert serializer._serialize_file(file) == expected


@pytest.mark.parametrize(
    "gpg_attr, gpg_value",
    [
        (' data-gpg-sig="true"', True),
        (' data-gpg-sig="false"', False),
        ('', None),
    ],
)
def test_serialize_file_html_gpg(gpg_attr: str, gpg_value: Optional[bool]) -> None:
    serializer = SerializerHtmlV1()

    file = File(
        filename="test.html",
        url="https://example.com/test.html",
        hashes={},
        gpg_sig=gpg_value,
    )
    expected = (
        '<a href="https://example.com/test.html"'
        f'{gpg_attr}'
        '>test.html</a><br/>\n'
    )
    assert serializer._serialize_file(file) == expected


def test_serialize_project_page_html() -> None:
    project_page = ProjectDetail(
        meta=Meta(api_version="1.0"),
        name="test-project",
        files=(
            File(filename="test.html", url="https://example.com/test.html", hashes={}),
            File(filename="test.txt", url="test.txt", hashes={}),
        ),
    )
    expected = """<!DOCTYPE html>
    <html>
    <head>
        <meta name="pypi:repository-version" content="1.0">
        <title>Links for test-project</title>
    </head>
    <body>
    <h1>Links for test-project</h1>
<a href="https://example.com/test.html">test.html</a><br/>
<a href="test.txt">test.txt</a><br/>
</body>
</html>"""
    serializer = SerializerHtmlV1()
    assert serializer.serialize_project_page(project_page) == expected


def test_serialize_project_list_html() -> None:
    project_list = ProjectList(
        meta=Meta(api_version="1.0"),
        projects=frozenset([
            ProjectListElement(name="test-project-1"),
            ProjectListElement(name="test-project-2"),
        ]),
    )
    expected_header = """<!DOCTYPE html>
    <html>
    <head>
        <meta name="pypi:repository-version" content="1.0">
        <title>Simple index</title>
    </head>
    <body>
"""
    expected_footer = """</body>
</html>"""

    serializer = SerializerHtmlV1()
    serialized_page = serializer.serialize_project_list(project_list)
    assert serialized_page.startswith(expected_header)
    assert serialized_page.endswith(expected_footer)
    a_tags = serialized_page.removeprefix(expected_header).removesuffix(expected_footer)
    assert '<a href="test-project-1/">test-project-1</a><br/>' in a_tags
    assert '<a href="test-project-2/">test-project-2</a><br/>' in a_tags


def test_serialize_project_page_json() -> None:
    page = ProjectDetail(
        Meta("1.0"),
        "project",
        files=(
            File(
                filename="test1.whl",
                url="test1.whl",
                hashes={},
            ),
            File(
                filename="test2.whl",
                url="test2.whl",
                hashes={"hash": "test_hash"},
                requires_python=">4.0",
                dist_info_metadata=True,
                yanked="yanked",
                gpg_sig=True,
            ),
            File(
                filename="test3.whl",
                url="test3.whl",
                hashes={},
                dist_info_metadata={"sha": "..."},
                yanked=True,
            ),
        ),
    )
    serializer = SerializerJsonV1()
    res = serializer.serialize_project_page(page)

    assert json.loads(res) == json.loads('''
        {
            "meta": {
                "api-version": "1.0"
            },
            "name": "project",
            "files": [
                {
                    "filename": "test1.whl",
                    "url": "test1.whl",
                    "hashes": {}
                },
                {
                    "filename": "test2.whl",
                    "url": "test2.whl",
                    "hashes": {"hash": "test_hash"},
                    "requires-python": ">4.0",
                    "core-metadata": true,
                    "yanked": "yanked",
                    "gpg-sig": true
                },
                {
                    "filename": "test3.whl",
                    "url": "test3.whl",
                    "hashes": {},
                    "core-metadata": {"sha": "..."},
                    "yanked": true
                }
            ]
        }
    ''')


@pytest.mark.parametrize(
    "version, serialization", [
        (
            "1.0",
            '''[{
                "filename": "test1-1.0.whl",
                "url": "test1.whl",
                "hashes": {}
            }]''',
        ),
        (
            "1.1",
            '''[{
                "filename": "test1-1.0.whl",
                "url": "test1.whl",
                "hashes": {},
                "size": 1,
                "upload-time": "2000-01-04T00:00:00Z"
            }]''',
        ),
    ],
)
def test_serialize_project_page_json__v1_1_attrs(
    version: str,
    serialization: str,
) -> None:
    page = ProjectDetail(
        Meta(version),
        "project",
        files=(
            File(
                filename="test1-1.0.whl",
                url="test1.whl",
                hashes={},
                size=1,
                upload_time=datetime(2000, 1, 4, 0, 0, 0),
            ),
        ),
    )
    serializer = SerializerJsonV1()
    res = serializer.serialize_project_page(page)
    assert json.loads(res)["files"] == json.loads(serialization)


def test_serialize_project_list_json() -> None:
    page = ProjectList(
        Meta("1.0"),
        projects=frozenset([
            ProjectListElement("a"),
        ]),
    )
    serializer = SerializerJsonV1()
    res = serializer.serialize_project_list(page)

    assert res == json.dumps(
        json.loads('''{
        "meta": {
            "api-version": "1.0"
        },
        "projects": [
            {"name": "a"}
        ]
    }'''),
    )



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

from datetime import datetime, timedelta
import logging
import os
import pathlib
from urllib.parse import quote_plus
import uuid

import httpx

from .http import HttpRepository

error_logger = logging.getLogger("gunicorn.error")


class CachedHttpRepository(HttpRepository):
    """
    Caches the http responses received from the source,
    manages cache invalidation using ETAGS.
    If the source is unavailable, will return, if available,
    cached elements (even if stale).
    """

    def __init__(
        self,
        url: str,
        cache_path: pathlib.Path,
        http_client: httpx.AsyncClient | None = None,
        connection_timeout: timedelta = timedelta(seconds=15),
    ):
        super().__init__(url, http_client, connection_timeout)
        self._cache_path = cache_path.resolve()
        self._tmp_dir = cache_path / ".incomplete"
        self._tmp_dir.mkdir(parents=True, exist_ok=True)

    def _get_from_cache(self, page_url: str) -> str | None:
        cached_resource_path = self._cache_path / quote_plus(page_url)
        if not cached_resource_path.is_file():
            return None
        now = datetime.now().timestamp()
        # Depending on the OS configuration, atime modification
        # on read may be disabled, so we set it explicitly.
        os.utime(cached_resource_path, (now, now))
        return cached_resource_path.read_text()

    def _save_to_cache(self, page_url: str, content: str) -> None:
        cached_resource_path = self._cache_path / quote_plus(page_url)
        timestamp = datetime.now().strftime("%Y%m%d%H%M%S")
        dest_file = self._tmp_dir / f"{timestamp}_{uuid.uuid4().hex}"
        dest_file.write_text(content)
        # Use rename atomicity to avoid set/get race conditions
        dest_file.rename(cached_resource_path)

    async def _fetch_simple_page(
        self,
        page_url: str,
    ) -> tuple[str, str]:
        """Retrieves a simple page from the given URL. The retrieved page,
        content type and etag are cached. If the cached content is
        unchanged or the source is unavailable, the cached data is returned.
        """
        headers = {"Accept": self.downstream_content_types}

        if cached_content := self._get_from_cache(page_url):
            etag, cached_content_type, cached_page = cached_content.split(",", 2)
            headers.update({"If-None-Match": etag})

        try:
            response = await self._http_client.get(
                url=page_url,
                headers=headers,
                timeout=self._connection_timeout.total_seconds(),
            )
        except httpx.HTTPError as e:
            # If the connection to the source fails, and there is a cached page for
            # the requested URL, return the cached content, otherwise raise the error.
            error_logger.error(
                f"Connection to {page_url} failed with"
                f" the following error: {str(e)}",
            )
            if cached_content:
                return cached_page, cached_content_type
            raise

        if response.status_code == 304 and cached_content:
            # The cached content is still valid.
            return cached_page, cached_content_type

        response.raise_for_status()

        body = response.text
        content_type = response.headers.get("Content-Type", "")
        if new_etag := response.headers.get("ETag", ""):
            # If the ETag is set, cache the response for future use.
            self._save_to_cache(page_url, ",".join([new_etag, content_type, body]))
        return body, content_type
